import math
import StringIO
import tempfile
import subprocess
import os
import os.path
import shutil
import geometry

class Painter:
    def __init__(self):
        pass

    def reset(self, x = None, y = None, layer = None):
        pass

    def strip(self, x, y):
        pass

    def via(self, layer):
        pass

    def pad(self, number):
        pass

    def save(self, filename):
        pass

    def push(self):
        pass

    def pop(self):
        pass

    def comment(self, blurb):
        pass

class Kicad(Painter):
    hole_min = .3
    ring_min = .1524

    layer_maps = [
        ["F.Cu", "B.Cu"],
        ["F.Cu", "B.Cu", "In1.Cu"],
        ["F.Cu", "In1.Cu", "B.Cu"],
        ["F.Cu", "In1.Cu", "In2.Cu", "B.Cu"],
    ]
    def __init__(self, ant):
        import kicad_mod as Ki

        self.stack = []
        self.layer_map = self.layer_maps[ant.layers - 1]

        self.module = Ki.Module(ant.filename(),
            "NFC antenna", "virtual")
        self.trace_width = ant.trace.width * 1000
        self.hole_diameter = max(self.trace_width / math.pi, self.hole_min)
        self.ring_diameter = max(self.trace_width,
                                 self.hole_diameter + self.ring_min * 2)

    def save(self, filename):
        if not filename:
            filename = self.module.name
        if not filename.endswith(".kicad_mod"):
            filename += ".kicad_mod"

        fd = open(filename, 'w')
        fd.write(self.module.code().encode("latin-1"))
        fd.close()

    def reset(self, x = None, y = None, layer = None):
        import kicad_mod as Ki

        if x is not None and y is not None:
            self.position = Ki.Point(x * 1000, y * 1000)
        if layer is not None:
            self.layer = layer

    def strip(self, x, y):
        import kicad_mod as Ki

        to = Ki.Point(x * 1000, y * 1000)

        if self.position == to:
            return

        import kicad_mod as Ki

        self.module.append(Ki.Line(self.position, to,
                                   self.trace_width, self.layer_map[self.layer]))

        self.position = to

    def via(self, layer):
        import kicad_mod as Ki

        self.module.append(Ki.Via(self.position, self.hole_diameter, self.ring_diameter))
        self.layer = layer

    def pad(self, number):
        import kicad_mod as Ki

        self.module.append(Ki.Connection(self.position, self.ring_diameter, number,
                                         self.layer_map[self.layer]))

    def push(self):
        self.stack.append((self.position, self.layer))

    def pop(self):
        self.position, self.layer = self.stack.pop()

    def comment(self, blurb):
        import kicad_mod as Ki

        self.module.append(Ki.Text(Ki.Point(0, 0), blurb,
                                   layer = "Cmts.User", font = Ki.Font.small))

class FastHenry(Painter):
    layer_maps = [
        [1.6e-3, 0],
        [1.6e-3, 0, 260e-6],
        [1.6e-3, 1333e-6, 0],
        [1.6e-3, 1333e-6, 260e-6, 0],
    ]

    def __init__(self, ant):
        self.ant = ant
        self.name = ant.filename()

        self.layer_map = self.layer_maps[ant.layers - 1]
        self.stack = []

        self.points = {}
        self.segments = set()
        self.equi = {}
        self.pads = {}

    def save(self, filename):
        buffer = StringIO.StringIO()
        buffer.write("** %s **\n" % self.name)
        buffer.write(".units m\n")
        buffer.write(".default w=%E h=%E sigma=%E\n" % (
            self.ant.trace.width, self.ant.trace.thickness,
            self.ant.trace.conductivity))

#        print buffer.getvalue()

        for (x, y, z), no in sorted(self.points.items(), lambda x, y: cmp(x[1], y[1])):
            buffer.write("N%d x=%E y=%E z=%E\n" % (no, x, y, z))

        for no, (a, b) in enumerate(sorted(self.segments)):
            buffer.write("E%d N%d N%d\n" % (no, a, b))

        for nodeset in self.equi.values():
            buffer.write(".equiv %s\n" % (" ".join(["N%d" % n for n in nodeset])))

        pads = list(sorted(self.pads.items()))
        buffer.write(".external N%d N%d P%d_%d\n" % (
            pads[0][1], pads[-1][1], pads[0][0], pads[-1][0]))
        buffer.write(".freq fmin=13.56e6 fmax=13.56e6\n")
        buffer.write(".end\n")

#        print buffer.getvalue()

        return self.run(buffer.getvalue())

    def run(self, data):

        directory = tempfile.mkdtemp()
        process = subprocess.Popen(["fasthenry", "-l", "auto", "-a", "on", "-"],
                                   stdin = subprocess.PIPE,
                                   stdout = subprocess.PIPE,
                                   cwd = directory)
        process.communicate(data)

        fd = open(os.path.join(directory, "Zc.mat"), "r")
#        print fd.read()
#        fd.seek(0)
        fd.readline()
        fd.readline()
        l = fd.readline().strip()
        fd.close()

        shutil.rmtree(directory)

        imag, real = l.split()
        Z = complex(float(imag), float(real[:-1]))
#        print l, Z

        return Z

    def point(self, x, y, z):
        try:
            return self.points[(x, y, z)]
        except:
            n = len(self.points) + 1
            self.points[(x, y, z)] = n
            return n

    def reset(self, x = None, y = None, layer = None):
        if x is not None and y is not None:
            self.position = x, y

        if layer is not None:
            self.layer = layer

        a = self.point(self.position[0], self.position[1], self.z())

        try:
            eq = self.equi[(self.position[0], self.position[1])]
            eq.add(a)
        except:
            pass

    def z(self):
        return self.layer_map[self.layer]

    def strip(self, x, y):
        z = self.z()

        if self.position == (x, y):
            return

        a = self.point(self.position[0], self.position[1], z)
        b = self.point(x, y, z)

        self.segments.add((a, b))
        self.position = x, y

    def via(self, layer):
        a = self.point(self.position[0], self.position[1], self.z())
        self.layer = layer
        b = self.point(self.position[0], self.position[1], self.z())

        k = (self.position[0], self.position[1])
        eq = self.equi.get(k, set())
        eq.add(a)
        eq.add(b)
        self.equi[k] = eq

    def pad(self, number):
        p = self.point(self.position[0], self.position[1], self.z())
        self.pads[number] = p

    def push(self):
        self.stack.append(self.position)

    def pop(self):
        self.position = self.stack.pop()

    def comment(self, blurb):
        pass

class FastHenryImage(FastHenry):
    def run(self, data):
        directory = tempfile.mkdtemp()
        process = subprocess.Popen(["fasthenry", "-f", "simple", "-"],
                                   stdin = subprocess.PIPE,
                                   stdout = subprocess.PIPE,
                                   cwd = directory)
        process.communicate(data)

        process = subprocess.Popen(["zbuf", "zbuffile"],
                                   stdin = subprocess.PIPE,
                                   stdout = subprocess.PIPE,
                                   cwd = directory)
        process.communicate()

        process = subprocess.Popen(["ps2pdf14", "-dEPSCrop",
                                    os.path.join(directory, "zbuffile.ps"),
                                    self.name + ".pdf"])
        process.communicate()

        shutil.rmtree(directory)

class FastCap(Painter):
    thickness = 1.6e-3
    mil = 1e-3 * 2.54e-2
    prepreg_thickness = 10 * mil
    mask_thickness = .5 * mil

    # Relative permittivity
    fr4_rp = 4.8
    mask_rp = 3.3
    air_rp = 1.

    # Hints on http://www.icd.com.au/articles/Stackup_Planning_AN2011_2.pdf

    draw_mask = True

    def __init__(self, ant):
        self.ant = ant
        self.name = ant.filename()

        self.layer = None
        self.position = 0, 0, 0

        self.peers = {}
        self.strips = []

        self.stack = []

        self.quads = [[], [], [], []]
        self.current_quad = None
        self.layer = 0
        self.position = None

    def reset(self, x = None, y = None, layer = None):
        self.current_quad = None
        if x is not None and y is not None:
            self.position = geometry.Point(x, y)
        if layer is not None:
            self.layer = layer

    def push(self):
        self.stack.append((self.position, self.layer))

    def pop(self):
        self.position, self.layer = self.stack.pop()

    def strip(self, x, y):
        dest = geometry.Point(x, y)
        
        if dest == self.position:
            return

        if self.position:
            l = geometry.Line(self.position, dest)
            q = geometry.Quad.from_segment(l, self.ant.trace.width)
            self.quads[self.layer].append(q)

            if self.current_quad:
                try:
                    self.current_quad.connect_to(q)
                except:
                    pass

            self.current_quad = q

        self.position = dest

    def via(self, layer):
        self.current_quad = None
        self.layer = layer

    def pad(self, number):
        pass

    def comment(self, blurb):
        pass

    def save(self, filename):
        directory = tempfile.mkdtemp()
        main = open(os.path.join(directory, "main.lst"), "w")
        main.write("G coil\n")

        t = self.ant.trace.thickness

        # z: altitude of bottom of conductor
        # material: on other side of panel [bottom, right, top, left]
        if self.ant.layers == 1:
            z = [self.thickness / 2 - self.mask_thickness - t,
                 -self.thickness / 2 + self.mask_thickness]
            material = [["fr4", "mask", "mask", "mask"],
                        ["mask", "mask", "fr4", "mask"],
                        ]
        elif self.ant.layers == 2:
            z = [self.thickness / 2 - self.mask_thickness - t,
                 -self.thickness / 2 + self.mask_thickness,
                 self.thickness / 2 - self.mask_thickness - t - self.prepreg_thickness - t]
            material = [["fr4", "mask", "mask", "mask"],
                        ["mask", "mask", "fr4", "mask"],
                        ["fr4", "fr4", "fr4", "fr4"],
                        ]
        elif self.ant.layers == 4:
            z = [self.thickness / 2 - self.mask_thickness - t,
                 self.thickness / 2 - self.mask_thickness - t - self.prepreg_thickness - t,
                 -self.thickness / 2 + self.mask_thickness + t + self.prepreg_thickness,
                 -self.thickness / 2 + self.mask_thickness]
            material = [["fr4", "mask", "mask", "mask"],
                        ["fr4", "fr4", "fr4", "fr4"],
                        ["fr4", "fr4", "fr4", "fr4"],
                        ["mask", "mask", "fr4", "mask"],
                        ]

        for layer, quads in enumerate(self.quads):
            if not quads:
                continue

            ml = material[layer]
            
            for i, q in enumerate(quads):
                center = q.center
                q0 = q - center

                files = {}

                for face, coords in enumerate([
                    (q0.a.x, q0.a.y, -t/2, q0.b.x, q0.b.y, -t/2, q0.c.x, q0.c.y, -t/2, q0.d.x, q0.d.y, -t/2),
                    (q0.b.x, q0.b.y, -t/2, q0.b.x, q0.b.y, t/2, q0.c.x, q0.c.y, t/2, q0.c.x, q0.c.y, -t/2),
                    (q0.a.x, q0.a.y, t/2, q0.b.x, q0.b.y, t/2, q0.c.x, q0.c.y, t/2, q0.d.x, q0.d.y, t/2),
                    (q0.a.x, q0.a.y, -t/2, q0.a.x, q0.a.y, t/2, q0.d.x, q0.d.y, t/2, q0.d.x, q0.d.y, -t/2),
                    ]):
                    m = ml[face]

                    if m not in files:
                        name = "strip-%d-%d-%s.cap" % (layer, i, m)
                        files[m] = open(os.path.join(directory, name), "w")
                        files[m].write("0 Strip %d on layer %d / %s\n" % (i, layer, m))
                    files[m].write("Q 1  %g %g %g  %g %g %g  %g %g %g  %g %g %g\n" % coords)

                for j, (m, f) in enumerate(files.items()):
                    name = "strip-%d-%d-%s.cap" % (layer, i, m)
                    main.write("C %s %f %g %g %g +\n" % (name, getattr(self, m + "_rp"), center.x, center.y, z[layer]))
                    f.close()

        main.seek(main.tell() - 3)
        main.write("\n")
        main.truncate()

        if self.draw_mask:
            mask = open(os.path.join(directory, "mask.cap"), "w")
            mask.write("0 solder mask\n")
            mask.write("Q 2  %g %g %g  %g %g %g  %g %g %g  %g %g %g\n" % (
                -self.ant.width, -self.ant.height, -self.thickness / 2,
                -self.ant.width, self.ant.height, -self.thickness / 2,
                self.ant.width, self.ant.height, -self.thickness / 2,
                self.ant.width, -self.ant.height, -self.thickness / 2,
                ))
            mask.write("Q 2  %g %g %g  %g %g %g  %g %g %g  %g %g %g\n" % (
                -self.ant.width, -self.ant.height, self.thickness / 2,
                -self.ant.width, self.ant.height, self.thickness / 2,
                self.ant.width, self.ant.height, self.thickness / 2,
                self.ant.width, -self.ant.height, self.thickness / 2,
                ))
            mask.close()

            main.write("G mask\n")
            main.write("D mask.cap %f %f  0 0 0  0 0 0\n" % ((1., self.mask_rp)))

        main.close()

        r = self.process(directory, "main.lst")

        shutil.rmtree(directory)

        return r

    def process(self, directory, main):
        process = subprocess.Popen(["fastcap", "-l" + main],
                                   stdout = subprocess.PIPE,
                                   cwd = directory)
        out, err = process.communicate()
        lines = filter(None, map(str.strip, out.split("\n")))
        last = lines[-1]
        cap = float(last.split()[-1])
        return cap * 1e-12

class FastCapImage(FastCap):
    draw_mask = False

    def process(self, directory, main):
        process = subprocess.Popen(["fastcap", "-m", "-l" + main],
                                   stdout = subprocess.PIPE,
                                   cwd = directory)
        process.communicate()

        process = subprocess.Popen(["ps2pdf14", "-dEPSCrop",
                                    os.path.join(directory, os.path.splitext(main)[0] + ".ps"),
                                    self.name + ".pdf"],
                                   stdout = subprocess.PIPE,
                                   )
        process.communicate()
